/****************************************************************************/

#include "winlib.h"

/*****************************************************************************
 * Name         - CDispatcher::subclass
 * Purpose      - When this is called, it will subclass the given procedure
 *                by storing the window procedure from the window into the
 *                m_OrgHandler member variable, then telling the window to
 *                use the passed in window procedure for all other messages.
 *                It also changes the handler type for this dispatcher to
 *                the subclass dispatcher.
 * Arguments    - owner: The handle to the window that owns this dispatcher,
 *                       which is the window that will be subclassed
 *                wndProc: Pointer to the window procedure to subclass the
 *                         current window with.
 * Return Value - None
 ****************************************************************************/
void CDispatcher::subclass (HWND owner , WNDPROC wndProc)
{
 WINASSERT (owner);
 WINASSERT (wndProc);

 orgHandler = (WNDPROC) GetWindowLong (owner , GWL_WNDPROC);
 SetWindowLong (owner , GWL_WNDPROC , (LONG) wndProc);
 handler = handlerSubclass;
}

/*****************************************************************************
 * Name         - CDispatcher::dispatchMessage
 * Purpose      - This function does the grunt work for the CDispatcher class.
 *                It checks first to see if the message to be dispatched is a
 *                command message. If it is, and the derived class has defined
 *                a command table, then the message is sent to the command
 *                dispatching function. Otherwise, it checks the message
 *                list (if any) to see if there is a matching handler. If it
 *                finds one it uses the given function to handle the message.
 *                otherwise, it calls the default handler to take care of the
 *                message.
 * Arguments    - Standard window callback parameters
 * Return Value - Depends on the message
 ****************************************************************************/
LRESULT CDispatcher::dispatchMessage (HWND hWnd , UINT uMessage ,
                                      WPARAM wParam , LPARAM lParam)
{
 // Is the message a command message? If it is, and there is a command list,
 // dispatch the message to the default handler.
 if (uMessage == WM_COMMAND && getCommandTable ())
     return dispatchCommand (hWnd , uMessage , wParam , lParam);

 // Otherwise, it must not be a command message. Do we have a command table?
 // If so, try to find the command to execute for it.
 MessageEntry *msg = getMessageTable ();
 if (msg) {
     for (INT index = 0 ; msg[index].pHandler != NULL ; index++)
         if (msg[index].uMessage == uMessage) {
             CDispatcher *dispatcher = this;
             PFNMSGHANDLER func = msg[index].pHandler;
             WINASSERT (func);
             return (dispatcher->*func) (hWnd , uMessage , wParam , lParam);
         }
 }

 // If we get here, there was either no message table, or there was a table
 // but no entry for this message. Call the default handler for this
 // function then.
 return defaultDispatch (hWnd , uMessage , wParam , lParam);
}

/*****************************************************************************
 * Name         - CDispatcher::dispatchCommand
 * Purpose      - This function is called by the main dispatch function to
 *                handle WM_COMMAND messages. It assumes there is a command
 *                table, because the dispatcher only calls it if there is one.
 *                It scans the command table to see if it can find a match,
 *                and if it can, it uses the proper function to handle the
 *                message. Otherwise, it sends the message for default
 *                handling.
 * Arguments    - Standard window callback parameters
 * Return Value - Depends on the command message
 ****************************************************************************/
LRESULT CDispatcher::dispatchCommand (HWND hWnd , UINT uMessage ,
                                      WPARAM wParam , LPARAM lParam)
{
 // This only gets called if the message received was a WM_COMMAND message
 // and there was a command table. So get a pointer to the table and then
 // see if we have an entry for this command or what.
 CommandEntry *cmd = getCommandTable ();
 WINASSERT (cmd);

 // Get the command ID for the command message
 WORD wCommand = GET_WM_COMMAND_ID (wParam , lParam);

 // Scan the table looking for a command ID that looks like one we have
 // a handler for. If we find one, call the handler and return what it
 // returns.
 for (INT index = 0 ; cmd[index].pHandler != NULL ; index++)
     if (cmd[index].wCommand == wCommand) {
         CDispatcher *dispatcher = this;
         PFNCMDHANDLER func = cmd[index].pHandler;
         WINASSERT (func);
         return (dispatcher->*func) (hWnd , wCommand ,
                                     GET_WM_COMMAND_CMD (wParam , lParam),
                                     GET_WM_COMMAND_HWND (wParam , lParam));
     }

 // If we get here, we don't have a defined handler for this command ID
 // so send the message for default handling instead
 return defaultDispatch (hWnd , uMessage , wParam , lParam);
}

/*****************************************************************************
 * Name         - CDispatcher::defaultDispatch
 * Purpose      - This function is called to handle all messages that do not
 *                have an explicit handler. Based on the dispatch class's
 *                handler type, it calls the proper function for default
 *                handling.
 * Arguments    - Standard window callback parameters
 * Return Value - Depends on the message
 ****************************************************************************/
LRESULT CDispatcher::defaultDispatch (HWND hWnd , UINT uMessage ,
                                      WPARAM wParam , LPARAM lParam)
{
 switch (handler) {
     case handlerNone:
         return 0;

     case handlerWindow:
         return DefWindowProc (hWnd , uMessage , wParam , lParam);

     case handlerDialog:
         return DefDlgProc (hWnd , uMessage , wParam , lParam);

     case handlerMDIFrame:
         WINASSERT (hwndMDIClient);
         return DefFrameProc (hWnd , hwndMDIClient , uMessage , wParam ,
                              lParam);

     case handlerMDIChild:
         return DefMDIChildProc (hWnd , uMessage , wParam , lParam);

     case handlerSubclass:
         WINASSERT (orgHandler);
         return CallWindowProc (orgHandler , hWnd , uMessage , wParam , lParam);
 }

 return 0;
}

/****************************************************************************/

