#ifndef __WINDISPATCHdotH__
#define __WINDISPATCHdotH__

/*
OK. Our dispatcher class by default has no command or message tables. In
addition it defines member functions to return a pointer to the message or
command table, which of course by default return NULL (meaning, no table).
The constructor for the base class is private, meaning that in order to use
the class, you must derive an instance of it. In the new derived instance,
you can use macros to declare a message dispatcher, a command dispatcher,
or both. There are these 4 possibilities:
1) No message or command tables
       - All calls to dispatch route to the proper default handling code
2) Message tables
       - Messages in the table get routed to their handler code
3) Command tables
       - WM_COMMAND messages get routed to the proper handler
4) Message and command tables
       - Any message in either table gets routed

The WM_COMMAND message gets treated specially, in that there doesn't have to
be an entry for it in the message dispatch table to dispatch commands. As long
as there is a command table, the dispatch code will handle it.

The Dispatcher can get all the information that it needs to handle dispatching
all messages on construction, except for an MDI Frame window default handling
and subclassed windows. For MDI Frame default processing to work, the class
must know the handle to the MDI Client window, but that is not available
on window construction. To get around that, specify default frame handling
for the dispatcher globally, then once the MDI Client window is created tell
the Dispatcher it's HWND, and it will store the HWND for later and change the
handler type to HandlerMDIFrame.

There is also the situation of a subclassed window. The Dispatch class must
be specified globally so it is in existance to handle the very first WM_CREATE
message. Also, you need the HWND from the WM_CREATE message to subclass the
window. The solution here is to once again set up some other sort of handler
by default in the Dispatcher, and then handle the WM_CREATE message for that
window and use the dispatcher's subclass routine to subclass the window and
change it's default handler. From that point on the window will be subclassed.
This is ok because 99.999% of the time a window that is being subclasses has
already been created, and the reason for subclassing it is simply to handle
a few specific messages.

Both of these situations work just fine because both of them stem from needing
an HWND to do their jobs, and the MWND becomes valid in the WM_CREATE message,
and since that is the first message sent to a window, there is no worry about
default processing since you can handle the WM_CREATE message and change the
default processing before it becomes an issue.
*/

/****************************************************************************/

#define MESSAGE_LIST \
  static MessageEntry msgTable[];    \
  virtual MessageEntry *getMessageTable (void)    \
      {return msgTable;}

#define START_MESSAGE_LIST(class) \
  MessageEntry class::msgTable[] = {

#define MESSAGE(msg,func) \
  {msg , (PFNMSGHANDLER) func},

#define END_MESSAGE_LIST \
  {0 , NULL }};

#define COMMAND_LIST \
  static CommandEntry cmdTable[];    \
  virtual CommandEntry *getCommandTable (void)    \
      {return cmdTable;}

#define START_COMMAND_LIST(class) \
  CommandEntry class::cmdTable[] = {

#define COMMAND(cmd,func) \
  {cmd , (PFNCMDHANDLER) func},

#define END_COMMAND_LIST \
  {0 , NULL}};

/****************************************************************************/

// This enumerated type is used to determine which of the default window
// procedure functions should be called by the message and command dispatching
// functions if a message or command is not handled.
typedef enum {
    handlerNone,      // Do not call any default procedure. (STD Dialogs use this)
    handlerWindow,    // Call DefWindowProc.
    handlerDialog,    // Call DefDlgProc (Custom dialogs only)
    handlerMDIChild,  // Call DefMDIChildProc.
    handlerMDIFrame,  // Call DefFrameProc.
    handlerSubclass   // The window is subclassed, so call the original WndProc
} handlerType;

// Forward declare the class so we can use it in a typedef
class CDispatcher;

// Function pointer prototype used for message handling functions and command
// handing functions.
typedef LRESULT (CDispatcher::*PFNMSGHANDLER) (HWND , UINT , WPARAM , LPARAM);
typedef LRESULT (CDispatcher::*PFNCMDHANDLER) (HWND , WORD , WORD , HWND);

// This structure maps a message to a message handling function.
typedef struct _MessageEntry {
    UINT          uMessage;
    PFNMSGHANDLER pHandler;
} MessageEntry;

// This structure maps a command ID to a command handling function
typedef struct _CommandEntry {
    WORD          wCommand;
    PFNCMDHANDLER pHandler;
} CommandEntry;

/****************************************************************************/

class CDispatcher {
protected:
    CDispatcher () : orgHandler (NULL) , handler (handlerNone) ,
        hwndMDIClient (NULL) {}
    CDispatcher (handlerType defHandler) : orgHandler (NULL) ,
        handler (defHandler) , hwndMDIClient (NULL) {}

public:
    void setMDIClient (HWND client) {
        hwndMDIClient = client;
        handler = handlerMDIFrame;
    }
    void subclass (HWND owner , WNDPROC wndProc);

    // Dispatch dispatches the message to the proper handler. If it is
    // a command message a command handler will be called otherwise a
    // message handler will be called. Default dispatches the message to
    // the default handling for a message. Where the message is sent from
    // there depends on the default handler type passed into the constructor.
    LRESULT dispatchMessage (HWND , UINT , WPARAM , LPARAM);
    LRESULT defaultDispatch  (HWND , UINT , WPARAM , LPARAM);

protected:
    virtual MessageEntry *getMessageTable (void) {return NULL;}
    virtual CommandEntry *getCommandTable (void) {return NULL;}

    LRESULT dispatchCommand (HWND , UINT , WPARAM , LPARAM);

    WNDPROC      orgHandler;
    handlerType  handler;
    HWND         hwndMDIClient;
};

/****************************************************************************/

#endif
